#!/usr/bin/env python
'''
 Copyright (C) 2011 - 2012 Andrew Hayzen

 This document conforms to the PEP 8 Style Guide

 This file is a python module called HTD (HayzenTech Database)

 Version 3.00 ALPHA-1

 This file is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This file is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with this file.  If not, see <http://www.gnu.org/licenses/>
'''

from types import ListType
import copy
import datetime
import hashlib
import os
import random
import re
import string
import struct
import sys
import time
import threading

if os.name == "posix":
    import fcntl
elif os.name == "nt":
    # FIXME: need windows file lock reading/writing
    import msvcrt
else:
    print("HTD Warning: Unknown operating system, FileLock unavailable.")


#--------------------------------------------------------------------- Settings
cache = True
debug = True
errors = True
force_sync = False
logging_hook = None
__version__ = 3.00


#------------------------------------------------------------------- Exceptions
class AuthenticationError(Exception):
    pass


class ConnectionError(Exception):
    pass


class VersionError(Exception):
    pass


#------------------------------------------------------------- Helper functions
def _convert_location(location, add):
    folder = location.split("/")
    return "%s/%s%s" % ("/".join(folder[:-1]), add, folder[-1])


#-------------------------------- Error Class - For managing and raising errors
class _Error():
    def __init__(self):
        self._error_config = {"error": errors, "debug": debug}
        self._logging_hook = logging_hook
        self._levels = ["Fatal", "Critical", "Warning"]

    def _raise(self, exctype, text=None, level=None):
        """ Raise an error """
        # 0 Fatal-Halt   1 Critical-DB State not expected   2 Warning-Continue
        if self._logging_hook is not None:
            self._logging_hook(exctype=exctype, text=text, level=level)

        if self._error_config["debug"]:
            if text is not None:
                if level is not None:
                    print('HTD %s: %s' % (self._levels[level], text))
                else:
                    print('HTD: %s %s' % (exctype, text))

        if self._error_config["error"]:
            if text is None:
                raise exctype()
            else:
                raise exctype(text)

    def set_error_settings(self, errors=True, debugging=True):
        """ Set error handling settings """
        self._error_config['error'] = errors
        self._error_config['debug'] = debugging
        return True


#----------------------------------------------------------- Thread lock object
class _ThreadLock():
    def __init__(self):
        self._threading_lock = threading.Lock()

    def _acquire(self):
        """ Acquire threading lock """
        return self._threading_lock.acquire()

    def _release(self):
        """ Release threading lock """
        return self._threading_lock.release()


#--------------------------------- File lock object - wraps around an open file
class _FileLock():
    def __init__(self, fileobj):
        self._locked_file = fileobj
        self._locked_location = fileobj.name
        self._lkcur = "UN"

    def _lock_reading(self):
        """ Lock file for reading """
        if os.name == "posix":
            fcntl.flock(self._locked_file, fcntl.LOCK_SH | fcntl.LOCK_NB)
        elif os.name == "nt":
            msvcrt.locking(self._locked_file.fileno(), msvcrt.LK_LOCK,
                               os.path.getsize(self._locked_location))
        self._lkcur = "SH"
        return

    def _lock_writing(self):
        """ Lock file for writing """
        if os.name == "posix":
            fcntl.flock(self._locked_file, fcntl.LOCK_EX | fcntl.LOCK_NB)
        elif os.name == "nt":
            msvcrt.locking(self._locked_file.fileno(), msvcrt.LK_LOCK,
                               os.path.getsize(self._locked_location))
        self._lkcur = "EX"
        return

    def _unlock(self):
        """ Unlock file """
        if os.name == "posix":
            fcntl.flock(self._locked_file, fcntl.LOCK_UN)
        elif os.name == "nt":
            try:
                msvcrt.locking(self._locked_file.fileno(), msvcrt.LK_UNLCK,
                               os.path.getsize(self._locked_location))
            except:
                pass
        self._lkcur = "UN"

        return


#-------------------------------------------------------------- Progress object
class _Progress():
    def __init__(self):
        self._tasks = {}

    def _count_prog(self, prog=1):
        """ Update progress of thread (internal only) """
        self._tasks[threading.currentThread()]['v'] += prog
        return

    def get_progress(self, thread):
        """ Get the details of a specific thread """
        if thread in self._tasks:
            try:
                t = copy.copy(self._tasks[thread])
                arr = {"value": t['v'], "length": t['l'], "status": t['s'],
                       "progress": float(t['v']) / float(t['l'])}
            except KeyError:
                arr = False
            finally:
                return arr
        else:
            return False

    def _set_prog(self, key, val):
        """ Set values for threading (internal only) """
        self._tasks[threading.currentThread()] = {'l': val, 'v': 0, 's': key}
        return True

    def _del_prog(self):
        """ Delete thread from list (internet only) """
        del self._tasks[threading.currentThread()]


#-------------------------------- Tools object - Holds generic helper functions
class _Tools():
    def __init__(self):
        # HTD config
        self._config = {"version": __version__}

        self._permissions = {"r": "read", "o": "overwrite", "a": "append",
                            "d": "delete", "x": "admin"}

    def _genericchecks(self, perm):
        """ Run generic connection and permission checks """
        if self._alive:
            if perm is None:
                return True
            for i in perm:
                if i not in self._access:
                    self._error(AuthenticationError,
                               'User doesn\'t have to permission (%s - %s)' %
                               (i, self._permissions[i]), 2)
                    return False
            return True
        else:
            self._error(ConnectionError, 'The connection has been closed.', 1)
            return False

    def _has_permission(self, username, password, location):
        """ Check username and password """
        auth = False
        user = self._header["USR"]
        if (username is None or not username and
            ("guest" in user and user["guest"][0] is None)):
            access = user["guest"][1]
            auth = True
        elif username in user:
            if password is not None:
                password = hashlib.sha256(password).hexdigest()
            if user[username][0] == password:
                access = user[username][1]
                auth = True

        # Raise auth error
        if not auth:
            self._file.unlock()
            self.close()
            self._error(AuthenticationError, 'Authentication Error for ' \
                       '(at %s).' % (location), 2)
            return False

        # Set user settings
        self._username = username
        self._access = access
        return True

    def _is_htd(self, location):
        # Check that file is HTD
        if not self._file._header_load():
            self._file.unlock()
            self.close()
            self._error(ConnectionError, 'Not a HTD (v%s) file (at %s) ' %
                       (self._config["version"], location), 1)
            return False

        # Check that file is same version
        if self._header["VER"] != self._config["version"]:
            self._file.unlock()
            self.close()
            self._error(VersionError, 'Version (v%s) of (at %s) is different' \
                        ' to HTD (v%s)' % (self._header["VER"], location,
                          self._config["version"]), 1)
            return False

        return True


#------------------------------- Header object - Manages the I/O for the header
class _Header():
    def __init__(self, new):
        if not new:
            self._header = self._header_load()
        else:
            self._header = {}

    def _get_atts_by_id(self, atts):
        """ Sort attributes by id """
        attributes = {}

        for att in atts:
            attr = atts[att]
            attributes[attr["id"]] = [attr["size"], att]

        return attributes

    def _header_load(self):
        """ Load header and return header """
        # Check data exists
        self._file.seek(0)
        data = self._file.read(8)
        if data == "":
            return False

        # Read data
        length = struct.unpack("Q", data)[0]
        pos = 8
        header = {}
        while pos < length:
            wid = struct.unpack("Q", self._file.read(8))[0]
            data = self._file.read(wid)
            header[data[0:3]] = data[3:]
            pos += 8 + wid

        # Convert data
        header["VER"] = float(header["HTD"])
        del header["HTD"]
        for att in header:
            if att != "VER":
                header[att] = eval(header[att])

        header["SIZ"] = length
        header["ATTBYID"] = self._get_atts_by_id(header["ATT"])

        # Update var
        self._header = header

        return self._header

    def _header_update(self, header, blank=False):
        """ Set new header """
        # Get old header
        if not blank:
            oldheader = copy.deepcopy(self._header)

        # Build inital str
        write = "%sHTD%s" % (struct.pack("Q", len(str(header["VER"])) + 3),
                              str(header["VER"]))

        # Add header atts
        for att in ["KEY", "USR", "TAB", "ATT"]:
            length = len(str(header[att])) + 3
            write += "%s%s%s" % (struct.pack("Q", length), att, header[att])

        # Add length
        write = "%s%s" % (struct.pack("Q", len(write) + 8), write)
        header["SIZ"] = len(write)

        # Write to file
        if blank:
            self[0:1] = write
        else:
            self[0:oldheader["SIZ"]] = write

        header["ATTBYID"] = self._get_atts_by_id(header["ATT"])

        self._header = header

        if not self._new:  # Skip if new
            self._cache_header_load()  # Update cache

        return header


#--------------------------------------- Index object - Holds the record widths
class _Index():
    def __init__(self, location, new):
        # Set vars
        self._index_cache = []
        self._index_cache_size = 50
        self._index_location = "%s.htd" % _convert_location(location,
                                                            ".~index_")

        # Create and connect to index
        if not os.path.exists(self._index_location) and not new:
            self._index_create_file(self._file, self._header["SIZ"],
                                    len(self._header["ATT"]))
        elif not new:
            self._index_file = open(self._index_location, "rb+")
            self._index_create_cache()

    def _index_close(self):
        """ Disconnect from index """
        if self._index_file is not None:
            self._index_file.close()
            self._index_file = None
        self._index_cache = []
        self._index_cache_size = 50
        return

    def _index_create_cache(self, cache=True):
        """ Create cache from connected file """
        # Calculate Cache Size
        length = self.get_length()
        self._index_cache_size = int((0.01 * length) + 50)

        # Work out how many items
        length = length / self._index_cache_size

        # Set vars
        x = 0
        output = []
        txt = "Q" * self._index_cache_size
        siz = self._index_cache_size * 8

        # Seek
        self._index_file.seek(0)

        # Read file
        while x < length:
            tmp = self._index_file.read(siz)
            output.append(sum(struct.unpack(txt, tmp)))
            x += 1

        # Set cache
        self._index_cache = output

        if cache:
            self._cache_index_load()  # Update cache

        return True

    def _index_create_file(self, fileobj, pos, numatts):
        """ Create index file from fileobj """
        # Create file
        f = open(self._index_location, "wb+")
        f.close()

        self._index_file = open(self._index_location, "rb+")

        self._index_clear()  # Clear cache

        # Get length of data and seek to start
        fileobj.seek(0, 2)
        end = fileobj.tell()
        fileobj.seek(pos)

        # Loop through data
        while pos < end:
            length = 0
            x = 0
            while x < numatts:
                data = fileobj.read(4)
                wid = struct.unpack("I", data)[0]
                fileobj.seek(wid, 1)
                wid += 4
                pos += wid
                length += wid
                x += 1
            self._index_file.write(struct.pack("Q", length))

        # Create cache
        self._index_create_cache()
        return True

    def _index_clear(self):
        """ Clear index """
        if self._index_file is not None:
            self._index_file.seek(0)
            self._index_file.truncate()
        self._index_cache = []
        return True

    def _index_delete(self):
        """ Delete index file """
        if os.path.exists(self._index_location) is False:
            return False

        self._index_cache = []
        os.unlink(self._index_location)
        return True

    def _index_delete_key(self, key):
        """ Delete a record from index """
        key *= 8
        length = (self.get_length() * 8) - (key + 8)
        self._index_file.seek(key + 8)
        text = self._index_file.read(length)
        self._index_file.seek(key)
        self._index_file.write(text)
        self._index_file.truncate()
        self._index_create_cache()
        return True

    def _index_insert(self, pos, value):
        """ Insert record into position """
        data = struct.pack("Q", value)
        pos *= 8
        length = (self.get_length() * 8) - pos
        self._index_file.seek(pos)
        text = self._index_file.read(length)
        text = '%s%s' % (data, text)
        self._index_file.seek(pos)
        self._index_file.write(text)
        self._index_file.truncate()
        self._index_create_cache()
        return True

    def _index_set(self, key, value):
        """ Change record width to value """
        self._index_file.seek(8 * key)
        data = struct.pack("Q", value)
        self._index_file.write(data)
        self._index_create_cache()
        return True

    def get_length(self):
        """ Get length (number of records) of index """
        self._index_file.seek(0, 2)
        length = self._index_file.tell() / 8
        return length

    def get_position(self, key):
        """ Get position in db of key """
        # Load cache
        cache = key / self._index_cache_size
        if cache == 0:
            length = 0
        else:
            length = sum(self._index_cache[0:cache])

        key = key % self._index_cache_size

        self._index_file.seek(cache * (self._index_cache_size * 8))

        # Read width
        data = self._index_file.read(key * 8)

        if key == 1:
            length += struct.unpack("Q", data)[0]
        elif key == 0:
            pass
        else:
            try:
                length += sum(struct.unpack("Q" * key, data))
            except struct.error:
                return False

        return length + self._header["SIZ"]

    def get_width(self, key):
        """ Get length of record """
        self._index_file.seek(8 * key)
        data = self._index_file.read(8)
        if data == '':
            return 0
        else:
            return struct.unpack("Q", data)[0]


#------ Cache object - Holds recent changes to index, header and a record cache
class _Cache():
    def __init__(self, location):
        self._cache_file = None

        self._cache_location = "%s.htd" % _convert_location(location,
                                                            ".~cache_")

        self.cache = cache

        # Open File
        if not os.path.exists(self._cache_location):
            f = open(self._cache_location, "wb")
            f.close()

        self._cache_file = open(self._cache_location, "rb+")
        self._cache_size = os.path.getsize(self._cache_location)

        # Set vars
        self._keys = {}
        self._lines = {}

    def add_line(self, line, rawdata, key):
        """ Add a cache item """
        if self.cache:
            self._lines[line] = [rawdata, key]
            self._keys[key] = line
        return

    def _cache_clear_line(self, line, fupdate=True):
        """ Clear cache items after line """
        # Loop through lines and delete if needed
        if line in self._lines:
            key = self._lines[line][1]
            if key in self._keys:
                del self._keys[key]
            del self._lines[line]

        # Update file
        if fupdate:
            self._cache_file.seek(0, 2)
            self._cache_file.write("\x15%s" % struct.pack("Q", line))
        return

    def _cache_clear_line_after(self, line, fupdate=True):
        """ Clear cache items after line """
        # Loop through lines and delete if needed
        for l in [l for l in self._lines if l >= line]:
            key = self._lines[l][1]
            if key in self._keys:
                del self._keys[key]
            del self._lines[l]

        # Update file
        if fupdate:
            self._cache_file.seek(0, 2)
            self._cache_file.write("\x16%s" % struct.pack("Q", line))

        return

    def _cache_clear_all(self, fupdate=True):
        """ Clear all cache items """
        self._keys = {}
        self._lines = {}

        # Update file
        if fupdate:
            self._cache_file.seek(0, 2)
            self._cache_file.write("\x17")
        return

    def _cache_close(self):
        """ Close cache """
        self._cache_file.close()
        self._cache_file = None

        self._keys = {}
        self._lines = {}
        return

    def _cache_confirm(self):
        """ Confirm cache to cache file """
        # Compare size
        length = os.path.getsize(self._cache_location)

        # Lock file and get locks
        while 1:
            try:
                self._lock_file._lock_writing()
                break
            except IOError:
                pass

        # Get locks
        locks = self._get_locks(False)

        position = 0

        # Loop through locks with ID
        for i, (o, p, c) in enumerate(locks):
            if o == self._id and p == os.getpid():  # Find self
                position = locks[i][2]   # Update position
                idc = i  # Store ID for later

        if position < length:
            f = open(self._cache_location)
            f.close()

            self._cache_file.seek(position)

            # Read any changes
            while position < length:
                self._cache_file.seek(position)
                data = self._cache_file.read(1)

                if data == "\x01":  # 01
                    self._header_load()  # Header reload
                elif data == "\x0b":  # 11
                    self._index_create_cache(False)  # Index reload cache
                elif data == "\x15":  # 21
                    # Clear a line from cache
                    data = self._cache_file.read(8)  # Get packed line
                    position += 8

                    # Clear line
                    data = struct.unpack("Q", data)
                    self._cache_clear_line(data, False)
                elif data == "\x16":  # 22
                    # Clear a line from cache and lines after
                    data = self._cache_file.read(8)  # Get packed line
                    position += 8

                    # Clear line
                    data = struct.unpack("Q", data)
                    self._cache_clear_line_after(data, False)
                elif data == "\x17":  # 23
                    self._cache_clear_all(False)  # Clear all lines from cache
                else:
                    pass

                position += 1

        elif length == 0:
            self._cache_size = 0

        # Update position
        locks[idc][2] = position

        # Find min
        pos = min([x for o, p, x in locks])

        # Update cache file
        f = self._cache_file
        f.seek(pos)
        tmp = f.read()
        f.seek(0)
        f.truncate()
        f.write(tmp)

        # Update lock file
        f = self._lock_file._locked_file
        f.seek(0)
        f.truncate()

        for o, p, c in locks:
            f.write(struct.pack("QII", o, p, c - pos))  # Take truncated

        self._lock_file._unlock()

        return

    def _cache_delete(self):
        """ Delete cache and disconnect """
        self._cache_close()

        # Delete file
        if os.path.exists(self._cache_location):
            os.unlink(self._cache_location)
        return

    def _cache_header_load(self):
        """ Header has changed """
        # Update file
        self._cache_file.seek(0, 2)
        self._cache_file.write("\x01")
        return

    def _cache_index_load(self):
        """ Index has changed """
        # Update file
        self._cache_file.seek(0, 2)
        self._cache_file.write("\x0b")
        return

    def get_line(self, line):
        """ Get cache item for line """
        if line in self._lines:
            return self._lines[line]
        else:
            return False

    def get_key(self, key):
        """ Get cache item for key """
        if key in self._keys:
            return self._keys[key]
        else:
            return False


#------------------------------------- File object - Manges I/O of the HTD file
class _File(_Header, _Index, _Cache, _FileLock, _ThreadLock):
    def __init__(self, location, new=False):
        """ Open file - new is true if no header, index, cache """
        if not os.path.exists:
            raise IOError("No HTD file found (%s.htd)" % location)

        # Load file
        self._file = open("%s.htd" % location, "rb+")
        self._new = new
        self._force_sync = force_sync

        # Add file lock
        _FileLock.__init__(self, self._file)

        # Load header, threadlock, cache, index
        _Header.__init__(self, new)
        _ThreadLock.__init__(self)

        if not new:
            _Cache.__init__(self, location)
            _Index.__init__(self, location, new)

        # Load lock
        lockloc = "%s.htd" % (_convert_location(location, ".~lock_"))
        if not os.path.exists(lockloc):
            f = open(lockloc, "wb")
            f.close()
        f = open(lockloc, "rb+")

        # Add and gain filelock
        self._lock_file = _FileLock(f)

        while 1:
            try:
                self._lock_file._lock_writing()
                break
            except IOError:
                pass

        # Update lock file
        self._lock_file._locked_file.seek(0, 2)

        # Generate ID's
        self._id = ''.join([random.choice([str(x) for x in range(10)])
                            for x in range(4)]) + str(time.time() * (10 ** 4))
        self._id = long(float(self._id))
        pid = os.getpid()

        if not new:
            cpos = self._cache_size
        else:
            cpos = 0
        self._lock_file._locked_file.write(struct.pack("QII", self._id, pid,
                                                       cpos))

        self._lock_file._locked_file.flush()
        self._lock_file._unlock()

        # Set current lock for file
        self._lock_cur = "UN"

    def __delitem__(self, key):
        if isinstance(key, slice):
            # Delete Slice
            self._file.seek(key.stop)
            text = self._file.read()
            self._file.seek(key.start)
            self._file.write(str(text))
            self._file.truncate()
            self._file.flush()
        else:
            # Delete record

            # Find pos
            pos = self.get_position(key)
            if pos is False:
                return False

            # Get width
            wid = self.get_width(key)

            # Delete from file
            del self[pos:pos + wid]

            # Update index and cache
            self._index_delete_key(key)
            self._cache_clear_line_after(key)
        return True

    def __getitem__(self, key):
        if isinstance(key, slice):
            # Return slice
            self._file.seek(key.start, 0)
            return self._file.read((key.stop - key.start))
        else:

            # Find pos
            pos = self.get_position(key)
            if pos is False:
                return False

            # Loop through attributes
            output = {}
            atts = self._header["ATTBYID"]

            self._file.seek(pos)

            # Read atts
            for att in atts:
                try:
                    wid = struct.unpack("I", self._file.read(4))[0]
                except struct.error:
                    output = False
                    break
                if wid == 0:
                    output[atts[att][1]] = None
                else:
                    output[atts[att][1]] = self._file.read(wid)

            # Return
            return output

    def __len__(self):
        self._file.seek(0, 2)
        return int(self._file.tell())

    def __setitem__(self, key, value):
        if isinstance(key, slice):
            # Slice
            i = key.start
            e = key.stop
            diff = e - i
            if (diff != value):
                self._file.seek(e)
                text = self._file.read()
                text = '%s%s' % (value, text)
                self._file.seek(i)
                self._file.write(text)
                self._file.truncate()
            else:
                self._file.seek(i)
                self._file.write(str(value))
            self._file.flush()
        else:
            # Write record
            # Record MUST exist
            pos = self.get_position(key)
            if pos is False:
                return False

            # Get width
            owid = self.get_width(key)

            # Loop through attributes
            string = ""
            atts = self._header["ATTBYID"]

            for att in atts:
                if value[atts[att][1]] is None:
                    string += "\x00" * 4
                else:
                    wid = struct.pack("I", len(str(value[atts[att][1]])))
                    string += "%s%s" % (wid, value[atts[att][1]])

            self[pos:pos + owid] = string

            # Update index
            wid = len(string)
            if owid != wid:
                self._index_set(key, wid)
                self._cache_clear_line_after(key)
            else:
                self._cache_clear_line(key)
        return

    def _get_locks(self, lock=True, arr=True):
        """ Return the locks on the cursor """

        if arr:
            out = []
        else:
            out = {}

        if lock:
            self._lock_file._lock_reading()

        f = self._lock_file._locked_file
        f.seek(0)

        while 1:
            data = f.read(16)
            if data == "":
                break

            o, p, c = struct.unpack("QII", data)

            if arr:
                out.append([o, p, c])
            else:
                out[(o, p)] = c

        if lock:
            self._lock_file._unlock()

        return out

    def close(self):
        """ Close file and sub classes """
        # Remove self from lock file
        while 1:
            try:
                self._lock_file._lock_writing()
                break
            except IOError:
                pass

        locks = self._get_locks(False)

        co = self._id
        cp = os.getpid()

        f = self._lock_file._locked_file

        idc = None
        for i, (o, p, c) in enumerate(locks):
            if o == co and p == cp:
                f.seek((i * 16) + 16)
                tmp = f.read()
                f.seek(i * 16)
                f.truncate()
                f.write(tmp)
                idc = i

        if idc is not None:
            del locks[idc]

        # Check for other locks
        if len(locks) == 0:
            # If no other locks
            os.remove(self._lock_file._locked_location)

            if not self._new:
                self._cache_delete()  # Delete and close cache

        else:
            # If locks exist
            if not self._new:
                self._cache_close()  # Close cache

        # Close lockf file
        self._lock_file._locked_file.close()

        if not self._new:
            self._index_close()  # Close index

        # Close file
        self._unlock()
        self._file.close()

        return True

    def insert_line(self, key, data):
        # Write record
        # Record MUST NOT exist
        pos = self.get_position(key)
        if pos is False:
            return False

        # Loop through attributes
        string = ""
        atts = self._header["ATTBYID"]

        for att in atts:
            if data[atts[att][1]] is None:
                string += "\x00" * 4
            else:
                wid = struct.pack("I", len(str(data[atts[att][1]])))
                string += "%s%s" % (wid, data[atts[att][1]])

        wid = len(string)
        self[pos:pos] = string

        # Update index
        self._index_insert(key, wid)
        self._cache_clear_line_after(key)
        return

    def lock_reading(self):
        self._acquire()

        while 1:
            try:
                self._lock_reading()
                break
            except IOError:
                pass

        self._lock_cur = "SH"

        self._file.flush()
        self._index_file.flush()
        self._cache_file.flush()
        self._lock_file._locked_file.flush()
        self._cache_confirm()  # Confirm cache, index, header

        return

    def lock_writing(self):
        self._acquire()

        while 1:
            try:
                self._lock_writing()
                break
            except IOError:
                pass

        self._lock_cur = "EX"

        self._file.flush()
        self._index_file.flush()
        self._cache_file.flush()
        self._lock_file._locked_file.flush()
        self._cache_confirm()  # Confirm cache, index, header

        return

    def unlock(self):
        self._file.flush()
        self._index_file.flush()
        self._cache_file.flush()
        self._lock_file._locked_file.flush()

        if self._force_sync and self._lock_cur == "EX":
            os.fsync(self._file)
            os.fsync(self._index_file)
            os.fsync(self._cache_file)

        try:
            self._unlock()
        except IOError:
            pass

        self._lock_cur = "UN"

        self._release()

        return


class _Authentication(_Error, _Tools):
    def __init__(self, parent, location, username, password):
        self._alive = True
        self._parent = parent
        self._parent._cursors.append(self)
        self._location = location

        # Open file
        self._file = _File(location, False)
        self._header = self._file._header

        _Error.__init__(self)
        _Tools.__init__(self)

        self._file.lock_reading()

        # Confirm file and permissions
        self._is_htd(location)
        self._has_permission(username, password, location)

        self._file.unlock()

    def __repr__(self):
        if self._alive:
            return ("HTD Authentication Object (%s@%s)" %
                    (self._username or "guest", self._location))
        else:
            return "HTD Closed Authentication Object"

    def close(self):
        """ Close the authentication """
        if self._alive:
            # Acquire lock and close auth
            self._alive = False
            # Remove from parent
            self._parent._cursors.remove(self)
            del self._parent
            # Close main file
            del self._header
            self._file.close()
            del self._file
        return

    def delete_user(self, username):
        """ Delete a user from the db """
        if not self._genericchecks('dx'):
            return False

        self._file.lock_writing()

        # Check username not self
        if username == self._username:
            self._file.unlock()
            self._raise(ValueError, "Cannot delete youself.", 2)
            return False

        # Check username exists
        if username not in self._header["USR"]:
            self._file.unlock()
            self._raise(KeyError, 'Could not find user to delete', 1)
            return False

        # Delete user and rebuild header
        header = self._header
        del header["USR"][username]
        self._file._header_update(header)
        self._file.unlock()
        return True

    def is_user(self, username):
        """ Return True is user exists in db """
        if not self._genericchecks('rx'):
            return False

        self._file.lock_reading()

        out = False
        if username in self._header["USR"]:
            out = True

        self._file.unlock()

        return out

    def list_users(self):
        """ List users and access level in db """
        if not self._genericchecks('rx'):
            return False

        self._file.lock_reading()

        output = {}
        users = self._header["USR"]
        for user in users:
            output[user] = users[user][1]

        self._file.unlock()

        return output

    def set_user(self, username, password, access="roadx"):
        """ Set user in settings for db """
        if not self._genericchecks("ax"):
            return False

        self._file.lock_writing()

        header = self._header
        for i in access:
            if i not in "roadx":
                self._file.unlock()
                self._raise(KeyError, 'Invalid permission for user (%s ' \
                           'for %s)' % (i, username), 1)
                return False

        password = hashlib.sha256(password).hexdigest()
        header["USR"][username] = [password, access]

        self._file._header_update(header)
        self._file.unlock()

        return True


class _Cursor(_Error, _Tools, _Progress):
    def __init__(self, parent, location, username=None, password=None):
        # Start cursor
        self._alive = True
        self._parent = parent
        self._parent._cursors.append(self)
        self._location = location

        # Open file
        self._file = _File(location, False)
        self._header = self._file._header

        # Start inherits
        _Error.__init__(self)
        _Tools.__init__(self)
        _Progress.__init__(self)

        # Types
        self._types = {'LONG': {'f': long, 'ex': ValueError, 'r': True},
                       'INTEGER': {'f': int, 'ex': ValueError, 'r': True},
                       'FLOAT': {'f': float, 'ex': ValueError, 'r': True},
                       'TEXT': {'f': str, 'ex': ValueError, 'r': False},
                        'UTF8': {'f': unicode, 'ex': ValueError, 'r': False},
                        'UTF16': {'f': unicode, 'ex': ValueError, 'r': False},
                        'UTF32': {'f': unicode, 'ex': ValueError, 'r': False},
                        'DATETIME': {'f': datetime.datetime, 'ex': ValueError,
                                     'r': False},
                        'TIME': {'f': datetime.time, 'ex': ValueError,
                                  'r': False},
                        'DATE': {'f': datetime.date, 'ex': ValueError,
                                 'r': False}
                    }

        self._unsan_types = {'INTEGER': int, 'LONG': long, 'FLOAT': float,
                             'DATETIME': self._string_datetime, 'TEXT': str,
                             'UTF8': self._string_utf8,
                             'UTF16': self._string_utf16,
                             'UTF32': self._string_utf32,
                             'TIME': self._string_time,
                             'DATE': self._string_date,
                             }

        # Confirm file and permissions
        self._is_htd(location)
        self._has_permission(username, password, location)

    def __repr__(self):
        if self._alive:
            return ("HTD Cursor Object (%s@%s)" %
                    (self._username or "guest", self._location))
        else:
            return "HTD Closed Cursor Object"

    def _get_line(self, key, new=False):
        """ Find line for key """
        if isinstance(key, str):  # Convert key to list
            key = [key]

        # Set up att widths, order
        atts = []
        attwidth = {}
        primarykey = []

        for att in self._header["KEY"]:
            atts.append(att)
            attwidth[att] = self._header["ATT"][att]['size']
            primarykey.append(str(key[att]))

        # Generate primarykey
        primarykey = "".join(primarykey)

        # Check for cache reference
        key = self._file.get_key(primarykey)
        if key is not False and new is False:
            cline = key
            datal, datap = self._file.get_line(cline)
        else:
            # Will take log(2, numrecords)
            # Calculate upper and lower
            size = self._file.get_length()
            div = size / (10 ** 2)
            if div == 0:
                div = 1
            upper = size
            lower = 0

            # Find line
            datal = None
            datap = []
            cline = upper / 2

            if size == 0:
                cline = False

            while size > 0:
                # Check if cline in index
                tmp = self._file.get_line(cline)

                if tmp is not False:
                    datal, datap = tmp
                else:
                    datal = self._file[cline]
                    if datal is False:  # Fix for end of file
                        break

                # Build data
                datap = [str(datal[att]) for att in atts]

                # Join, add to cache
                datap = "".join(datap)

                self._file.add_line(cline, datal, datap)

                # Check data not primarykey
                if datap == primarykey:
                    if new:
                        cline = False
                    break
                elif datap == '' and new:
                    break

                # Compare
                comp = (datap > primarykey) - (datap < primarykey)

                # Calc movement
                # upper 10 lower 0 cline 5
                if comp > 0:
                    upper = cline
                    # upper 1 lower 0 cline 1 => 0
                    if (upper - lower) == 1:
                        cline = lower
                        continue
                    else:
                        # upper 5 lower 0 cline 5 => 2.5
                        cline = lower + long((upper - lower) / 2)
                elif comp < 0:
                    # upper 1 lower 0 cline 0 => 1
                    if (upper - lower) == 1:
                        cline = upper
                        lower = upper
                        continue
                    else:
                        # upper 10 lower 0 cline 5 => 7.5
                        lower = cline
                        cline = lower + long((upper - lower) / 2)
                else:
                    if new:
                        cline = False
                    break
                if upper == lower:
                    if not new:
                        cline = False
                    break

        # Return result
        if new:
            return [cline]
        else:
            return [cline, datal]

    def _rebuild_data(self, mode, att):
        # Get length
        length = self._file.get_length()
        if length > 0:
            if mode == "ADD":
                task = "Set Attribute"
            else:
                task = "Delete Attribute"
            task += ", %s" % (att)

            x = 0

            self._set_prog(task, length)

            read = self._file._file.read
            self._file._file.seek(self._header["SIZ"])
            attributes = self._header["ATTBYID"]
            write = ""

            while x < length:
                for att in attributes:
                    if mode == "REM":
                        datatmp = read(4)
                        wid = struct.unpack("I", datatmp)[0]

                    if attributes[att][1] == att:
                        if mode == "REM":
                            read(wid)
                        else:
                            write += "\x00\x00\x00\x00"
                    else:
                        if mode == "ADD":
                            datatmp = read(4)
                            wid = struct.unpack("I", datatmp)[0]
                        write += datatmp + read(wid)
                x += 1

                self._count_prog()

            self._file._file.seek(self._header["SIZ"])
            self._file._file.write(write)
            self._file._file.truncate()

            self._del_prog()

            self._file._cache_clear_all()
            self._file._index_clear()

            numatts = len(self._header["ATT"])
            if mode == "REM":
                numatts -= 1
            self._file._index_create_file(self._file._file,
                                          self._header["SIZ"], numatts)
        return True

    def _sanitise(self, data):
        """ Check that data passes attribute settings """

        if data is None:
            return None

        err = False
        for att in data:
            tmp = data[att]
            # Check attribute exists
            if att not in self._header["ATT"]:
                err = (AttributeError, 'Failed to find sanitise settings for' \
                       ' attribute %s' % (att), 0)

            output = self._header["ATT"][att]

            if data[att] is None:
                continue

            # Filter Check
            if (output["filter"] is not None):
                x = re.findall(filter, tmp)
                if x[0] != tmp:
                    err = (ValueError, 'Filter failed to match pattern for ' \
                           '"%s".' % att, 2)

            typed = output["type"]

            # Type check
            if not isinstance(tmp, self._types[typed]['f']):
                name = ('%s, failed to  match preset pattern for %s. (%s)' %
                        (att, typed, tmp))
                err = (self._types[typed]['ex'], name, 2)

            # Convert data
            if typed in ["DATETIME", "TIME", "DATE"]:
                tmp = tmp.isoformat()

            if typed in ["UTF8", "UTF16", "UTF32"]:
                tmp = self._unicode_string(tmp, int(typed[3:]))

            # Length Check
            if (len(str(tmp)) > output["size"]):
                err = (ValueError, 'String length check failed for %s.' %
                       name, 2)

            # Range check
            if self._types[typed]['r'] and 'range' in output:
                if not (output["range"][1] < tmp < output["range"][0]):
                    err = (ValueError, '"%s" failed to match range for "%s"' \
                           '.' % (name, typed), 2)

            # List check
            if output["list"][0] and tmp not in output["list"][1]:
                err = (KeyError, '"%s" failed to match an item in list.' %
                       name, 2)

            data[att] = tmp

        if not err:
            return data
        else:
            apply(self._raise, err)
            self._file.unlock()
            return False

    def _string_date(self, rawstr):
        data = rawstr.split("-")
        return datetime.date(int(data[0]), int(data[1]), int(data[2]))

    def _string_datetime(self, rawstr):
        data = rawstr.split(".")
        dt = datetime.datetime.strptime(data[0], "%Y-%m-%dT%H:%M:%S")
        us = int(data[1].rstrip("Z"), 10)
        return dt + datetime.timedelta(microseconds=us)

    def _string_time(self, rawstr):
        data = rawstr.split(":")
        return datetime.time(int(data[0]), int(data[1]), int(data[2]))

    def _string_utf8(self, rawstr):
        return rawstr.decode("utf8")

    def _string_utf16(self, rawstr):
        return rawstr.decode("utf16")

    def _string_utf32(self, rawstr):
        return rawstr.decode("utf32")

    def _unicode_string(self, rawstr, mode):
        return rawstr.encode("utf%s" % mode)

    def _unsanitise(self, data):
        for att in data:
            typed = self._header["ATT"][att]["type"]
            if data[att] is not None:
                data[att] = self._unsan_types[typed.upper()](data[att])

        return data

    def close(self):
        """ Close the cursor """
        if self._alive:
            # Acquire lock, close cursor
            self._file.lock_writing()
            self._alive = False

            # Remove cursor in parent
            self._parent._cursors.remove(self)
            del self._parent

            # Clear variables
            del self._unsan_types

            # Close the file
            self._file.unlock()
            del self._header
            self._file.close()
            del self._file

        return

    def delete_attribute(self, attribute):
        """ Delete attribute and data from db """
        if not self._genericchecks('d'):
            return False

        self._file.lock_writing()

        # Confirm att exists
        if attribute not in self._header["ATT"]:
            self._file.unlock()
            self._raise(KeyError, 'Could not find attribute to delete. ' \
                       '(%s)' % (attribute), 1)
            return False

        self._rebuild_data("REM", attribute)

        # Delete attribute, rebuild header
        header = self._header
        del header["ATT"][attribute]
        self._file._header_update(header)

        self._file.unlock()
        return True

    def delete_record(self, key):
        """ Delete a record from the db """
        if not self._genericchecks('d'):
            return False

        self._file.lock_writing()

        # San data
        if len(self._header["KEY"]) == 1:
            key = [key]

        key = self._sanitise(dict(zip(self._header["KEY"], key)))

        # Check if record exists and get line
        line = self._get_line(key)

        if line[0] is False:
            self._file.unlock()
            self._raise(KeyError, ('No record, %s, was found to ' \
                                  'delete.' % (key)), 1)
            return False

        # Calc pos and rewrite data/index/cache
        del self._file[line[0]]

        self._file.unlock()
        return True

    def delete_table(self):
        """ Delete all data in the table """
        if not self._genericchecks('d'):
            return False

        self._file.lock_writing()

        # Seek to start table, truncate
        self._file._file.seek(self._header["SIZ"])
        self._file._file.truncate()

        # Clear cache/index
        self._file._cache_clear_all()
        self._file._index_clear()

        self._file.unlock()
        return True

    def get_attribute(self, attribute):
        """ Return attributes settings """
        if not self._genericchecks('r'):
            return False

        self._file.lock_reading()

        attributes = self._header["ATT"]
        if attribute in attributes:
            self._file.unlock()
            return attributes[attribute]
        else:
            self._file.unlock()
            self._raise(KeyError, 'Failed to find sanitise settings for ' \
                       'attribute (%s)' % (attribute), 1)
            return False

    def get_identifier(self):
        """ Return identifier for db """
        if not self._genericchecks('r'):
            return False

        # Lock file
        self._file.lock_reading()

        # Get data
        identifier = self._header["KEY"]

        # Unlock
        self._file.unlock()

        return identifier

    def get_length(self):
        """ Return the number of records in db """
        if not self._genericchecks('r'):
            return False

        # Lock file
        self._file.lock_reading()

        # Get length
        length = self._file.get_length()

        # Unlock
        self._file.unlock()

        return length

    def get_record(self, key):
        """ Return dict of the record from db """
        if not self._genericchecks('r'):
            return False

        # Set vars
        self._file.lock_reading()

        # Filter primarykey and get line
        if len(self._header["KEY"]) == 1:
            key = [key]

        key = self._sanitise(dict(zip(self._header["KEY"], key)))

        # Check key exists
        pos, data = self._get_line(key)
        if pos is False:
            self._file._unlock()
            self._raise(KeyError, 'No key, "%s", found' % key, 1)
            return False

        data = self._unsanitise(data)

        # Release locks
        self._file.unlock()

        return data

    def get_table(self, start=0, stop=-1):
        """ Return table as dict within limits """
        # Limit 0 = all, +x x from start
        if not self._genericchecks('r'):
            return False

        # Set vars
        output = {}
        read = self._file._file.read
        self._file.lock_reading()

        length = self._file.get_length()
        attributes = self._header["ATTBYID"]

        # Skip to data
        self._file._file.seek(self._header["SIZ"])

        if stop != -1:
            length = stop

        x = 0
        if start != 0:
            x = start
            self._file._file.seek(self._file.get_position(x))

        composite = False
        if len(self._header["KEY"]) > 1:
            composite = True
        else:
            key = self._header["KEY"][0]

        self._set_prog("Get Table, start=%s, stop=%s" % (start, stop), length)

        while x < length:
            tmp = {}
            for att in attributes:
                wid = struct.unpack("I", read(4))[0]
                if wid == 0:
                    tmp[attributes[att][1]] = None
                else:
                    tmp[attributes[att][1]] = read(wid)

            tmp = self._unsanitise(tmp)

            if composite:
                prime = tuple([tmp[attr] for attr in self._header["KEY"]])
            else:
                prime = tmp[key]

            output[prime] = tmp
            x += 1
            self._count_prog()

        self._del_prog()
        self._file.unlock()
        return output

    def get_table_version(self):
        """ Return version of db file """
        if not self._genericchecks('r'):
            return False

        # Lock file
        self._file.lock_reading()

        # Get version
        version = self._header["VER"]

        # Unlock
        self._file.unlock()

        return version

    def get_version(self):
        """ Return engine version """
        return self._config["version"]

    def is_attribute(self, attribute):
        """ Return True if attribute exists in db """
        if not self._genericchecks('r'):
            return False

        self._file.lock_reading()

        out = False
        if attribute in self._header["ATT"]:
            out = True

        self._file.unlock()
        return out

    def is_record(self, key):
        """ Return True if record exists in db """
        if not self._genericchecks('r'):
            return False

        # Lock file
        self._file.lock_reading()

        # Filter primarykey and get line
        if len(self._header["KEY"]) == 1:
            key = [key]

        key = self._sanitise(dict(zip(self._header["KEY"], key)))
        line = self._get_line(key)

        # Unlock
        self._file.unlock()

        # Return True if key exists
        if line[0] is False:
            return False
        else:
            return True

    def list_attributes(self):
        """ List attributes in db by id """
        if not self._genericchecks('r'):
            return False

        self._file.lock_reading()

        attsbyid = self._header['ATTBYID']
        output = [attsbyid[attribute][1] for attribute in attsbyid]

        self._file.unlock()
        return output

    def list_records(self, start=0, stop=-1):
        """ List the keys of all records """
        if not self._genericchecks('r'):
            return False

        key = self._header["KEY"]
        if len(key) > 1:
            output = apply(zip, [self.list_select_attribute(att, start, stop)
                                 for att in key])
        else:
            output = self.list_select_attribute(key[0], start, stop)
        return output

    def list_select_attribute(self, attribute, start=0, stop=-1):
        """ List data from a specific attribute """
        if not self._genericchecks('r'):
            return False

        output = []

        read = self._file._file.read
        self._file.lock_reading()

        # Skip to data
        self._file._file.seek(self._header["SIZ"])

        if attribute not in self._header["ATT"]:
            self._file.unlock()
            self._raise(KeyError, 'Could not find attribute to list. ' \
                       '(%s)' % (attribute), 1)
            return False

        length = self._file.get_length()

        if stop != -1:
            length = stop

        x = 0
        if start != 0:
            x = start
            self._file._file.seek(self._file.get_position(x))

        attributes = self._header["ATTBYID"]

        self._set_prog("List Select %s" % (attribute), length)
        while x < length:
            tmp = {}
            for att in attributes:
                d = read(4)

                wid = struct.unpack("I", d)[0]

                if attributes[att][1] == attribute:
                    if wid == 0:
                        tmp[attribute] = None
                    else:
                        tmp[attribute] = read(wid)
                else:
                    read(wid)
            tmp = self._unsanitise(tmp)

            output.append(tmp[attribute])
            x += 1
            self._count_prog()

        self._del_prog()
        self._file.unlock()

        return output

    def set_attribute(self, attribute, attributetype, size=None,
                      attributerange=None, attributefilter=None,
                      limitlist=False, required=False):
        """ Set attribute settings in db """
        if not self._genericchecks(''):
            return False

        self._file.lock_writing()

        if attribute in self._header["ATT"]:
            if not self._genericchecks('o'):
                self._file.unlock()
                return False
        else:
            if not self._genericchecks('a'):
                self._file.unlock()
                return False
        data = {}

        data["type"] = attributetype.upper()
        if data["type"] in ["INTEGER", "LONG", "FLOAT"]:
            if attributerange is None:
                data["range"] = [sys.maxint, -sys.maxint - 1]
            else:
                if (isinstance(attributerange, ListType) and
                    len(attributerange) == 2):
                    if (attributerange[0] > sys.maxint or
                        attributerange[1] < -sys.maxint - 1):
                        self._file.unlock()
                        self._raise(ValueError, 'Range value incorrect ' \
                                   'for %s.(%s)' % (data["type"], attribute),
                                   1)
                        return False
                else:
                    self._file.unlock()
                    self._raise(ValueError, 'Range is invalid (%s)' %
                               (attribute), 1)
                    return False
                data["range"] = attributerange

        if attribute in self._header["ATT"]:
            data["id"] = self._header["ATT"][attribute]["id"]
        else:
            data["id"] = len(self._header["ATT"])

        if limitlist is False:
            data["list"] = [False, []]
        else:
            data["list"] = [True, limitlist]

        data["size"] = size
        if attribute in self._header["ATT"]:
            oldsize = self._header["ATT"][attribute]["size"]
            if size is None:
                data["size"] = oldsize
        else:
            if size is None:
                data["size"] = 4294967295

        tmp = {"DATETIME": 26, "DATE": 10, "TIME": 8}

        if data["type"] in tmp:
            data["size"] = tmp[data["type"]]
        data["required"] = required

        data["filter"] = attributefilter

        if required and self._ifile.get_length() > 0:
            self._file.unlock()
            self._raise(ValueError, "Table must be empty to set an " \
                       "attribute if it is required. (%s)" % (attribute), 1)
            return False

        header = self._header
        header["ATT"][attribute] = data

        self._file._header_update(header)

        self._rebuild_data("ADD", attribute)

        self._file.unlock()
        return True

    def set_identifier(self, key):
        """ Set identifier for a db """
        if not self._genericchecks('o'):
            return False

        self._file.lock_writing()

        if (isinstance(key, str) or isinstance(key, int) or
            isinstance(key, long)):
            key = [str(key)]

        if self._file.get_length() > 0:
            self._raise(ValueError, "The table must be empty before " \
                       "changing the identifier.", 1)
            self._file.unlock()
            return False

        header = self._header
        header["KEY"] = key
        self._file._header_update(header)

        self._file.unlock()
        return True

    def set_record(self, key, data={}):
        """ Set record in db """
        if not self._genericchecks(None):
            return False

        # Set vars
        self._file.lock_writing()

        # Filter primarykey
        if len(self._header["KEY"]) == 1:
            key = [key]
        key = self._sanitise(dict(zip(self._header["KEY"], key)))

        # Check if record exists and get current att
        datatmp = {}
        pos = self._get_line(key, True)

        # If record exists get current record
        if pos[0] is False:
            pos2 = self._get_line(key)
            if pos2[0] is not False:
                datatmp = pos2[1]

        # Run filters
        data = self._sanitise(data)

        # Confirm data
        for att in self._header["ATT"]:
            if att not in data and att in datatmp:
                data[att] = datatmp[att]
            elif att not in data and att not in datatmp:
                if self._header["ATT"][att]["required"] is False:
                    data[att] = None
                else:
                    self._file._unlock()
                    self._error(ValueError, "Required attribute (%s) was " \
                                "NULL" % (att), 1)
                    return False

        # Add identifier to data
        for att in self._header["KEY"]:
            data[att] = key[att]

        # Write to file
        if pos[0] is False and self._genericchecks('o'):
            self._file[pos2[0]] = data
        else:
            if self._genericchecks('a'):
                self._file.insert_line(pos[0], data)

        self._file.unlock()
        return True

    def set_table(self, data):
        """ Set many records in db (assumes table is empty) """

        # data {"key" : {"att":"data","att":"data"}, "key2" : {"att":...}}
        if not self._genericchecks('a'):
            return False

        # Set vars
        self._file.lock_writing()

        key = self._header["KEY"]
        # Check for composite
        if len(key) > 1:
            composite = True
        else:
            composite = False

        # Sort data
        data = data.items()
        sdata = sorted(data, key=lambda k: [str(k[0])])
        attributes = self._header["ATTBYID"]
        self._file._file.seek(self._header["SIZ"])

        # Check that the table is empty
        if self._file.get_length() != 0:
            self._raise(ValueError, "The table must be empty before using" \
                       " set_table.", 1)
            self._file.unlock()
            return False

        x = 0

        self._set_prog("Create Table", len(sdata))

        for line in sdata:
            tmp = []

            datadict = {}
            for att in attributes:
                attr = attributes[att]

                # Convert composite
                if composite:
                    for num, att in enumerate(self._header["KEY"]):
                        line[1][att] = line[0][num]
                else:
                    line[1][self._header["KEY"][0]] = line[0]

                # Check if att exists
                if attr[1] not in line[1]:
                    line[1][attr[1]] = None

                # Check if NULL
                if line[1][attr[1]] is None:
                    if attr[1] in key:
                        # Error in NULL
                        self._file.unlock()
                        self._raise(ValueError, "Primary key " \
                                   "cannot be a NULL value", 1)
                        return False
                    elif self._header["ATT"][attr[1]]["required"]:
                        # Error if required
                        self._file.unlock()
                        self._raise(ValueError, "%s is required" %
                                   attr[1], 1)
                        return False
                    else:
                        datadict[attr[1]] = None
                else:
                    datadict[attr[1]] = line[1][attr[1]]

            # Sanitise
            datadict = self._sanitise(datadict)

            # Convert dict to list
            tmp = [datadict[attributes[att][1]] for att in attributes]

            # Write to file
            string = ""
            for i in tmp:
                if i is None:
                    string += "\x00" * 4
                else:
                    string += struct.pack("I", len(str(i))) + str(i)
            self._file._file.write(string)
            x += 1

            self._count_prog()

        self._del_prog()

        self._file._cache_clear_all()
        self._file._index_create_file(self._file._file, self._header["SIZ"],
                                   len(self._header["ATT"]))
        self._file.unlock()
        return True

    def update_table(self, data):
        """ Merges data and current db """
        if not self._genericchecks('road'):
            return False

        datatowrite = self.get_table()
        for item in data:
            datatowrite[item] = data[item]

        self.delete_table()

        self.set_table(datatowrite)
        return True


class _Connection(_ThreadLock, _Error, _Tools):
    def __init__(self, location, username=None, password=None):
        # Load subclasses
        _ThreadLock.__init__(self)
        _Error.__init__(self)
        _Tools.__init__(self)

        # Set vars
        self._alive = True
        self._cursors = []
        self._location = location
        self._username = username
        self._password = password
        self._isdb = False

        # Load tmp cursor - confirm HTD file and auth
        cur = self.cursor(internal=True)

        # Set access settings
        self._access = cur._access

        # Set _isdb
        if not cur._file._header["TAB"]:
            self._isdb = True

        cur.close()

    def __repr__(self):
        if self._alive:
            if self._isdb:
                return ("HTD DB Connection Object (%s@%s)" %
                        (self._username or "guest", self._location))
            else:
                return ("HTD Table Connection Object (%s@%s)" %
                        (self._username or "guest", self._location))
        else:
            return "HTD Closed Connection Object"

    def auth(self, table=None, internal=False):
        """ Create a new authentication object """
        if not self._genericchecks("x"):
            return False

        self._acquire()

        if (table is None or internal is True):
            auth = _Authentication(self, self._location, self._username,
                                   self._password)
            self._release()
            return auth
        else:
            # Get table data from db
            cur = self.cursor(internal=True)
            tbd = cur.get_record(table)
            cur.close()
            auth = _Authentication(self, tbd['location'], tbd['username'],
                                   tbd['password'])
            self._release()
            return auth

    def close(self):
        """ Close the connection """
        if not self._genericchecks(""):
            return False

        # Close cursors
        for cur in copy.copy(self.cursors()):
            cur.close()

        self._acquire()
        self._alive = False
        self._release()
        return

    def create_table(self, tablename, name, location, identifier,
                     username=None, password=None):
        """ If db - Create a new table in the db """
        if not self._genericchecks("xa"):
            return False

        self._acquire()
        if self._isdb:
            # Create table
            con = create_table(name, location, identifier, username, password)
            self._release()

            # Join table with parent
            self.join(tablename, con, None, username, password)
            con.close()
            return True
        else:
            self._release()
            return False

    def cursor(self, table=None, username=None, password=None,
               internal=False):
        """ Create a new cursor """
        if not self._genericchecks(""):
            return False

        if not internal:
            self._acquire()

        if (table is None or internal is True):

            cursor = _Cursor(self, self._location, username or self._username,
                             password or self._password)
            if not internal:
                self._release()
            return cursor
        else:
            # Get table data from db
            cur = self.cursor(internal=True)
            tbd = cur.get_record(table)
            cur.close()
            cursor = _Cursor(self, tbd['location'], username or
                             tbd['username'], password or tbd['password'])
            if not internal:
                self._release()
            return cursor

    def cursors(self):
        """ Return list of open cursors """
        if not self._genericchecks(""):
            return False

        self._acquire()
        for cur in copy.copy(self._cursors):
            if cur is None:
                self._cursors.remove(cur)
            else:
                if not cur._alive:
                    del self._cursors._parent
                    self._cursors.remove(cur)
        self._release()
        return self._cursors

    def delete(self, tables=False):
        """ Delete connection db (and tables) and disconnect """
        if not self._genericchecks("d"):
            return False

        locations = [self._location]

        # Close cursors
        for cur in copy.copy(self.cursors()):
            cur.close()

        # Load tables
        if tables and self._isdb:
            tablelocs = self.tables()
            for table in tablelocs:
                locations.append(tablelocs[table])

        self._acquire()

        for location in locations:
            # FIXME: No file lock acquired?
            # Delete main db file
            if os.path.exists("%s.htd" % location):
                os.unlink("%s.htd" % location)
            # Calc index file name
            folder = location.split("/")
            index_location = "%s/%s%s.htd" % ("/".join(folder[:-1]),
                                                    ".~index_", folder[-1])
            # Delete index file
            if os.path.exists(index_location):
                os.unlink(index_location)

        self._release()

        self.close()
        return True

    def is_db(self):
        """ Return if the connected is a db """
        if not self._alive:
            return None

        return self._isdb

    def join(self, name, child, table=None, username=None,
             password=None):
        """ Join table with parent """
        if not self._genericchecks("xa"):
            return False

        self._acquire()

        # Get location and create data
        if table is None:
            loc = child._location
        else:
            locs = child.tables()
            if table in locs:
                loc = locs[table]
            else:
                self._release()
                self.errors(KeyError, "No table found in child, %s" % table)
                return False

        data = {"username": username, "password": password, "location": loc}

        # Open cursor on parent, Write data and name, close cursor
        cur = self.cursor(internal=True)
        cur.set_record(name, data)
        cur.close()
        self._release()
        return True

    def set_table_auth(self, name, username=None, password=None):
        """ Set the username and password for a child table """
        if not self._genericchecks("xo") or not self._isdb:
            return False

        self._acquire()

        data = {"username": username, "password": password}

        # Open cursor on parent, Write data and name, close cursor
        cur = self.cursor(internal=True)

        if not cur.is_record(name):
            self._release()
            self.errors(KeyError, "No child table found in parent, %s" % name)
            return False

        cur.set_record(name, data)

        cur.close()

        self._release()
        return True

    def split(self, name, delete=False):
        """ Split table with parent with the option to delete the table """
        if not self._genericchecks("xd"):
            return False

        self._acquire()

        # Open cursor, remove table, close
        cur = self.cursor(internal=True)

        if cur.is_record(name):
            data = cur.get_record(name)  # Get auth details
            cur.delete_record(name)  # Delete table reference
            cur.close()

            self._release()

            if delete:  # Delete table
                con = connect(data['location'], data['username'],
                              data['password'])
                con.delete()
            return True
        else:
            self._release()
            self.errors(KeyError, "No table to split named '%s'." % name)
            return False

    def tables(self):
        """ If a db - return the tables """
        # Check connections is DB
        if not self._genericchecks("r") or not self._isdb:
            return False

        self._acquire()

        # Open cursor, list table, close
        cur = self.cursor(internal=True)

        output = {}
        for tbl in cur.list_records():
            output[tbl] = cur.get_record(tbl)["location"]

        cur.close()
        self._release()

        return output


def create_db(name, location, username=None, password=None):
    """ Create a new db and return connection object """
    # Create db file
    con = create_table(name, location, "alias", username, password, False)

    # Open cursor
    cur = con.cursor()

    # Add attributes
    for att in ["alias", "location", "username", "password"]:
        cur.set_attribute(att, "TEXT")

    # Close and return con obj
    cur.close()
    return con


def create_table(name, location, identifier, username=None, password=None,
                 table=True):
    """ Create a new table and return connection object """
    # Check for file
    if os.path.exists("%s/%s.htd" % (location, name)):
        err = _Error()
        err._raise(IOError, 'File already exists for (%s at %s.)' %
                   (name, location), 1)
        return False

    # Load parameters
    if isinstance(identifier, str):
        identifier = [identifier]

    if username is None:
        username = "guest"

    passwordhash = None
    if password is not None:
        passwordhash = hashlib.sha256(password).hexdigest()

    user = {username: [passwordhash, "roadx"]}
    header = {"TAB": table, "ATT": {}, "USR": user, "KEY": identifier,
              "VER": __version__}
    location = "%s/%s" % (location, name)

    # Create file
    f = open("%s.htd" % location, "w")
    f.close()

    # Load and set header
    f = _File(location, True)
    f._header_update(header, True)
    f.close()

    # Return con obj
    return _Connection(location, username, password)


def connect(location, username=None, password=None):
    """ Returns a connection to a table """
    if os.path.exists("%s.htd" % location):
        return _Connection(location, username, password)
    else:
        err = _Error()
        err._raise(ConnectionError, "No HTD file exists at (%s)." %
                   location, 1)
        return False
